import numpy as np
import math
import pandas as pd
import os
import pickle
import copy
import sys
from scipy.spatial.distance import cdist
from scipy.spatial.transform import Rotation

class swcToolkit():
	"""
	
	```
	from roots.swcToolkit import swcToolkit
	swctool = swcToolkit()
	morph1 = swctool.load(morph1_filename)
	morph2 = swctool.load(morph2_filename)
	morph3 = swctool.combine_morphologies(morph1,morph2)
	swctool.to_swc(morph3,'morph3.swc')
	```
	
	"""
	def __init__(self):
		pass
	
	def shift_point(self,a,vec):
		"""
		
		Shift a point around an origin according to user specified vector
		
		a - point to be shifted
		vector - x,y,z values to shift point by
		
		returns the shifted point
		
		"""
		
		return([a[0]-vec[0],a[1]-vec[1],a[2]-vec[2]])
	
	def move_morphology(self,arbor,vector): #unit - microns
		"""
		
		Shift a morphology around an origin according to user specified vector
		
		arbor - dictionary with branch number as keys and sections as values (each section is a list of points)
		vector - x,y,z values to shift all points by
		
		returns the shifted morphology
		
		"""
		if any(isinstance(i, list) for i in arbor[0]):
			for branch in arbor.keys():
				for section in arbor[branch]:
					arbor[branch][arbor[branch].index(section)] = [self.shift_point(point[:3],vector)+[point[-1]] for point in section]
		
		else:
			for branch in arbor.keys():
				arbor[branch]= [self.shift_point(point[:3],vector)+[point[-1]] for point in arbor[branch]]
		
		return(arbor)
	
	def rotate_points(self,origin, points, ax=0.0,ay=0.0,az=0.0):
		"""
		
		Rotate a point counterclockwise by a given angle around a given origin.
		The angle should be given in radians.
		
		"""
		r = Rotation.from_rotvec(np.array([ax,ay,az]))
		rotated_points = np.array(points) - np.array(origin)
		rotated_points = r.apply(rotated_points)
		return(rotated_points+np.array(origin))
	
	def rotate_morphology(self,arbor,origin,about_x=0.0,about_y=0.0,about_z=0.0): 
		"""
		
		Rotate a morphology around an origin according to user specified elevation and azimuth
		
		arbor - dictionary with branch number as keys and sections as values (each section is a list of points)
		origin - point about which to rotate the morphology
		elevation - x,y plane in degrees
		azimuth - x,z plane in degrees
		
		returns the rotated morphology
		
		"""
		newarbor = {}
		pts = []
		if any(isinstance(i, list) for i in arbor[0]):
			for branch in arbor.keys():
				for section in arbor[branch]:
					for pt in section:
						pts.append(pt[:3])
			
			rotated_points = self.rotate_points(origin,np.array(pts),ax=np.pi*about_x/180.0,ay=np.pi*about_y/180.0,az=np.pi*about_z/180.0)
			pt_index = 0
			for branch in arbor.keys():
				newarbor[branch] = []
				for section in arbor[branch]:
					newarbor[branch].append([])
					for pt in section:
						newarbor[branch][-1].append(list(rotated_points[pt_index])+[pt[3]])
						pt_index+=1
		
		else:
			for branch in arbor.keys():
				for pt in arbor[branch]:
					pts.append(pt[:3])
			
			rotated_points = self.rotate_points(origin,np.array(pts),ax=np.pi*about_x/180.0,ay=np.pi*about_y/180.0,az=np.pi*about_z/180.0)
			pt_index = 0
			for branch in arbor.keys():
				newarbor[branch] = []
				for pt in arbor[branch]:
					newarbor[branch].append(list(rotated_points[pt_index])+[pt[3]])
					pt_index+=1
		
		
		return(newarbor)
	
	
	def split_branch_at_bifurcation(self,branch,b_point):
			return(branch[:branch.index(b_point)],branch[branch.index(b_point):])
	
	def reorder_arbor(self,arbor,a,b,c):
		newarbor = {}
		newbranch_count = 0
		for branch in arbor.keys():
			if arbor[branch] == a:
				newarbor[newbranch_count] = b
				newbranch_count+=1
				newarbor[newbranch_count] = c
				newbranch_count+=1
			else:
				newarbor[newbranch_count] = arbor[branch]
				newbranch_count+=1
		
		return(newarbor)
	
	def find_branch_w_bifurcation_point(self,arbor,b_point):
		for branch in arbor.keys():
			if b_point in arbor[branch][1:-1]:
				print(branch)
				return(branch)
		
		print(None)
		return(None)
	
	def split_branches_at_bifurcations(self,arbor):
		b_points = [arbor[b][0] for b in arbor.keys()]
#		new_arbor = copy.deepcopy(arbor)
		for point in b_points:
			a = self.find_branch_w_bifurcation_point(arbor,point)
			if a is not None:
				a = arbor[a]
			
			else:
				continue
			
			b,c = self.split_branch_at_bifurcation(a,point)
			arbor = self.reorder_arbor(arbor,a,b,c)
		return(arbor)
	
	def prepend_headers(self,filename):
		with open(filename,'r',encoding='utf-8') as f:
			dat = f.readlines()
			if dat[0] == 'n T x y z R P\n':
				return()
		with open(filename,'w',encoding='utf-8') as f:
			f.write('n T x y z R P\n')
			for line in dat:
				f.write(line)
				
	
	def cut_headers(self,filename):
		with open(filename,'r',encoding='utf-8') as f:
			dat = f.readlines()
		if dat[0] == 'n T x y z R P\n':			
			with open(filename,'w',encoding='utf-8') as f:
				for line in dat[1:]:
					f.write(line)
			return()
		else:
			return()
	
	def sort_index(self,li):
		newli = sorted(li)
		return([li.index(l) for l in newli])
	
	def find_children(self,bif,np_dict):
		children = []
		for n in np_dict.keys():
			if np_dict[n] == bif:
				children.append(n)
		return(children)
	
	def load(self,fname,asTree=True,interbifurcated=True):
		self.prepend_headers(fname)
		dat = pd.read_table(fname,delim_whitespace=True)
		self.cut_headers(fname)
		for col in ['n','T','P']:
			dat[col] = dat[col].astype('int')
		for col in ['x','y','z','R']:
			dat[col] = dat[col].astype('float')
	
		
		if asTree:
			if not interbifurcated:
				arbor = {}
				n_point = dict(zip(dat['n'],zip(dat['x'],dat['y'],dat['z'],dat['R'])))
				np_dict = dict(zip(dat['n'],dat['P']))
				branchnumber = 0
				arbor[0] = [n_point[1]]
				for n in np_dict.keys():
					if np_dict[n] == -1:
						continue
					if np_dict[n] != n-1:
						branchnumber+=1
						arbor[branchnumber] = [n_point[np_dict[n]]]
					
					arbor[branchnumber].append(n_point[n])
				
				
				return(arbor)
			
			else:
				arbor = {}
				n_point = dict(zip(dat['n'],zip(dat['x'],dat['y'],dat['z'],dat['R'])))
				np_dict = dict(zip(dat['n'],dat['P']))
				branchnumber = 0
				arbor[0] = [n_point[1]]
				for n in np_dict.keys():
					if np_dict[n] == -1:
						continue
					if np_dict[n] != n-1:
						branchnumber+=1
						arbor[branchnumber] = [np_dict[n]]
					
					arbor[branchnumber].append(n)
				
				terminals = [arbor[b][-1] for b in arbor.keys()]
				
				arbor = {}
				pn_dict = dict(zip(dat['P'],dat['n']))
				p_point = dict(zip(dat['P'],zip(dat['x'],dat['y'],dat['z'],dat['R'])))
				branchnumber = 0
				bifrows = [1]
				for n in np_dict.keys():
					if n == 1:
						continue
					if np_dict[n] != n-1:
						bifrows.append(np_dict[n])
				
				bifrows.sort()
				children = {}
				for b in bifrows:
					children[b] = self.find_children(b,np_dict)
				
				for childs in children.values():
					for child in childs:
						arbor[branchnumber] = [np_dict[child],child]
						branchnumber+=1
				
				for n,row in enumerate(dat['n']):
					if row-1 not in terminals and row-1 not in bifrows:
						for branch in arbor.keys():
							if np_dict[row] == arbor[branch][-1]:
								arbor[branch].append(row)
								break
				
				for branch in arbor.keys():
					arbor[branch] = [n_point[n] for n in arbor[branch]]
				
				
				return(arbor)
		
		else:
			x = {0:[dat['x'][0]]}
			y = {0:[dat['y'][0]]}
			z = {0:[dat['z'][0]]} 
			r = {0:[dat['R'][0]]}
			
			
			for p,point in enumerate(dat['x'][1:]):
				inc = dat['P'][p+1]
				if inc != dat['n'][p]:
					bnum = len(x.keys())
					x[bnum] = [dat.query('n == '+str(dat['P'][p+1]))['x'].values[0],dat['x'][p+1]]
					y[bnum] = [dat.query('n == '+str(dat['P'][p+1]))['y'].values[0],dat['y'][p+1]]
					z[bnum] = [dat.query('n == '+str(dat['P'][p+1]))['z'].values[0],dat['z'][p+1]]
					r[bnum] = [dat.query('n == '+str(dat['P'][p+1]))['R'].values[0],dat['R'][p+1]]
					continue
				else:
					bnum = len(x.keys())-1
					x[bnum].append(dat['x'][p+1])
					y[bnum].append(dat['y'][p+1])
					z[bnum].append(dat['z'][p+1])
					r[bnum].append(dat['R'][p+1])
			return(x,y,z,r)
	
	
	def to_swc(self,arbor=[],target='swcTest.swc'):
		
		"""
		n T x y z R P
		
		n is an integer label that identifies the current point and increments by one from one line to the next.
		
		T is an integer representing the type of neuronal segment, such as soma, axon, apical dendrite, etc. The standard accepted integer values are given below.
		
			0 = undefined
			1 = soma
			2 = axon
			3 = dendrite
			4 = apical dendrite
			5 = fork point
			6 = end point
			7 = custom
		
		x, y, z gives the cartesian coordinates of each node.
		
		R is the radius at that node.
		P indicates the parent (the integer label) of the current point or -1 to indicate an origin (soma). 
		"""
		
		try:
			with open(target,'wb') as f:
				points = [arbor[0][0]]
				n = [1]
				T = 2
				R = 0.7
				P = [-1]
				nind = 2
				Pind = 1
				f.write(str(n[-1])+' '+str(T)+' '+str(points[0][0])+' '+str(points[0][1])+' '+str(points[0][2])+' '+str(R)+' '+str(P[-1]))
				f.write('\n')
				for key in arbor.keys():
					for p,point in enumerate(arbor[key][1:]):
						points.append(point)
						n.append(nind)
						if p == 0 and key is not 0:
							P.append(points.index(arbor[key][0])+1)
						else:
							P.append(Pind)
						nind+=1
						Pind+=1
						f.write(str(n[-1])+' '+str(T)+' '+str(point[0])+' '+str(point[1])+' '+str(point[2])+' '+str(R)+' '+str(P[-1]))
						f.write('\n')
		
		except:
			with open(target,'w',encoding='utf-8') as f:
				points = [arbor[0][0]]
				n = [1]
				T = 2
				R = 0.1
				P = [-1]
				nind = 2
				Pind = 1
				f.write(str(n[-1])+' '+str(T)+' '+str(points[0][0])+' '+str(points[0][1])+' '+str(points[0][2])+' '+str(R)+' '+str(P[-1]))
				f.write('\n')
				for key in arbor.keys():
					for p,point in enumerate(arbor[key][1:]):
						points.append(point)
						n.append(nind)
						if p == 0 and key is not 0:
							P.append(points.index(arbor[key][0])+1)
						else:
							P.append(Pind)
						nind+=1
						Pind+=1
						f.write(str(n[-1])+' '+str(T)+' '+str(point[0])+' '+str(point[1])+' '+str(point[2])+' '+str(R)+' '+str(P[-1]))
						f.write('\n')
	
	def find_best_merge_point(self,a,b):
		points_a = []
		for branch in a.keys():
			points_a+=a[branch]
		
		points_a = np.array([np.array(item[:3]) for item in points_a]).reshape(len(points_a,3))
		
		points_b = []
		for branch in b.keys():
			points_b+=b[branch]
		
		points_b = np.array([np.array(item[:3]) for item in points_b]).reshape(len(points_b,3))
		
		distances = cdist(points_a,points_b)
		return(points_a[distances.argmin()])
	
	
	
	def combine_morphologies(self,a,b,merge_point=None):
		if merge_point == None:
			b[0] = self.find_best_merge_point(a,b)+b[0]
		else:
			b[0] = merge_point+b[0]
		
		for branchnumber in b.keys():
			a[len(a.keys())+branchnumber] = b[branchnumber]
		
		return(a)





from scipy.interpolate import interp1d
import numpy as np
import math
import os
import random
from shapely.geometry import LineString, MultiPoint
from shapely.ops import split
from shapely import wkt
from scipy.spatial.distance import cdist

class Microstructures():
	"""
	A class to add microstructures such as boutons and myelin to an axon morphology
	
	```
	from roots.microstructures import Microstructures
	mstruct = Microstructures()
	mstruct.primarytest()
	arbor = {}
	arbor[0] = [[0,0,0],[1,0,0],[2,0,0]]
	mstruct.add_microstructures_to_arbor(arbor,mbranches,bbranches)
	```
	
	"""
	
	def __init__(self):
		pass
	
	def apply_microstructures_diameter_scalars(self,arbor,arbor_labels,label_scalars={},replace=True):
		if label_scalars == {}:
			print('didn\'t do any scaling because label_scalars=={}')
			return(arbor)
		
		if replace:
			for branch in arbor.keys():
				for s,section in enumerate(arbor[branch]):
					arbor[branch][s] = [list(item[:-1])+[label_scalars[arbor_labels[branch][s]]] for item in arbor[branch][s]]
		
		else:
			for branch in arbor.keys():
				for s,section in enumerate(arbor[branch]):
					arbor[branch][s] = [list(item[:-1])+[label_scalars[arbor_labels[branch][s]]*item[-1]] for item in arbor[branch][s]]
		
		return(arbor)
	
	def eucdist3d(self,point1,point2):
		"""
		
		euclidean distance between point1 and point2 - [x,y,z]
		
		"""
#		if not isinstance(point1,np.ndarray):
#			point1 = np.array(point1)
#			point2 = np.array(point2)
		
		return(((point2[0]-point1[0])**2 + (point2[1]-point1[1])**2 + (point2[2]-point1[2])**2)**0.5)
	
	def branchLength(self,branch):
		"""
		
		branch - list of points comprising a branch
		returns - the path length of the branch
		
		
		"""
		
		length = 0
		for p,point in enumerate(branch[:-1]):
			length+=self.eucdist3d(branch[p],branch[p+1])
		
		return(length)
	
	def parse_spline(self,shapelystring):
		st = wkt.loads(shapelystring)
#		sl = np.array(st[0])
#		if len(st) > 1:
#			for s in st[1:]:
#				sl = np.append(sl,np.array(s),axis=0)
	
		return([list(point) for point in np.array(st)])
	
	def reassign_diameters(self,original,new):
		complist = [p[:3] for p in original]
		for n,point in enumerate(new):
			new[n].append(original[np.argmin(cdist([list(point)],complist))][-1])
		
		return(new)
	
	def spline_branch(self,branch,interval=1.0):
		"""
		
		Spline branch to evenly spread points to make re-sectioning by myelin or boutons easier
		
		branch - list of points comprising a branch
		interval - maximum distance between points in newly splined branch
		
		"""
		
		number_of_points = int(round(self.branchLength(branch)/interval,0))+1
		line = LineString([tuple(point[:3]) for point in branch])
		splitter = MultiPoint([line.interpolate((i/number_of_points),normalized=True) for i in range(number_of_points+1)])
		interp = self.parse_spline(splitter.wkt)
		#interp = [list(branch[0][:3])] + interp[1:] + [list(branch[-1][:3])]
		return(self.reassign_diameters(branch,interp))
#		
#		x = [pnt[0] for pnt in branch]
#		y = [pnt[1] for pnt in branch]
#		z = [pnt[2] for pnt in branch]
#		r = [pnt[3] for pnt in branch]
#		newx = np.linspace(0,len(x)-1,num=number_of_points,endpoint=True)
#		newy = np.linspace(0,len(y)-1,num=number_of_points,endpoint=True)
#		newz = np.linspace(0,len(z)-1,num=number_of_points,endpoint=True)
#		newr = np.linspace(0,len(r)-1,num=number_of_points,endpoint=True)
#		if len(x) < 4:
#			if len(x) < 3:
#				order = 'linear'
#			
#			else:
#				order = 'quadratic'
#		else:
#			order = 'cubic'
#		
#		fx = interp1d(range(len(x)),x,kind=order)
#		fy = interp1d(range(len(y)),y,kind=order)
#		fz = interp1d(range(len(z)),z,kind=order)
#		fr = interp1d(range(len(r)),r,kind=order)
#		return(list(zip(fx(newx),fy(newy),fz(newz),fr(newr))))
	
	def find_optimal_segmentation_length(self,blength,seg_range):
		"""
		
		blength - length of the branch in microns
		seg_range - acceptable range of lengths for node+paranode1+paranode2+internode+paranode2+paranode1 length (as a unit of myelination) or interbouton+bouton length (as a unit of bouton size)
		returns - a list of segment lengths that optimally divides a branch into myelinated/boutoned regions
		
		"""
		seg_range = self.generate_seg_range(seg_range)
		easiestfit = seg_range[0]
		remainder = np.max(seg_range)
		for segr in seg_range:
			remains = blength - math.floor(blength/segr)*segr
			if remains<remainder:
				remainder=remains
				easiestfit=segr
		
		seg_lengths = [easiestfit for i in range(math.floor(blength/easiestfit))]
		return([segment+remainder/len(seg_lengths) for segment in seg_lengths])
	
	def divide_microstructure_unit(self,point1,point2,dimensions):
		"""
		
		point1 - beginning of region to be resegmented
		point2 - end of region to be resegmented
		dimensions - lengths to cut region
		returns - new segments that make up the microstructure being added to the region
		
		"""
		new_sections = []
		if len(point1) < 4:
			done = 0.0
			dtwo = 0.0
		else:
			done = point1[-1]
			dtwo = point2[-1]
		
		p1 = np.array(point1[:3])
		p2 = np.array(point2[:3])
		vec = p2-p1
		dimslength = float(np.sum(dimensions))
		for d,dim in enumerate(dimensions[:-1]):
			nearsideproportion = np.sum(dimensions[:d])/dimslength
			farsideproportion = np.sum(dimensions[:d+1])/dimslength
			new_sections.append([	
						list(np.append(p1+vec*nearsideproportion,done)),
						list(np.append(((p1+vec*nearsideproportion)+(p1+vec*farsideproportion))/2.0,(done+dtwo)/2.0)),
						list(np.append(p1+vec*farsideproportion,dtwo))
						])
		
		new_sections.append([	
					list(new_sections[-1][-1]),
					list((np.array(new_sections[-1][-1])+np.array(list(point2[:3])+[dtwo]))/2.0),
					list(point2[:3])+[dtwo]
					])
		
		if len(dimensions) > 2:
			return(new_sections,['node','paranode1','paranode2','internode','paranode2','paranode1'][:len(new_sections)])
		
		else:
			return(new_sections,['interbouton','bouton'][:len(new_sections)])
	
	def generate_seg_range(self,length):
		return(np.arange(length*0.75,length*1.25,0.1))
	
	def myelinate_branch(self,branch,myelin_dimensions):
		"""
		This function finds optimal microstructure dimensions to fit within a user specified branch, 
		redistributes points through the branch corresponding to the breakpoitns of these section,
		and returns the end points of each of these new structures.
		
		branch - list of points comprising a branch
		myelin_dimensions - descriptions of node, paranode1,paranode2, and internode lengths (microns)
		returns - branch where list members are end points of each new section
		
		"""
		
		try:
			segments = self.find_optimal_segmentation_length(self.branchLength(branch),seg_range=np.sum(myelin_dimensions))
		except:
			return(self.segment_and_label_branch(branch,'node'))
		
		if segments == []:
			return(self.segment_and_label_branch(branch,'node'))
		
		try:
			splined_branch = self.spline_branch(branch,interval=segments[0])
		except:
			return(self.segment_and_label_branch(branch,'node'))
		
		segmented_branch = []
		labellist = []
		for s,segment in enumerate(splined_branch[:-1]):
			microstructure,labels=self.divide_microstructure_unit(splined_branch[s],splined_branch[s+1],myelin_dimensions[:3]+[self.branchLength(splined_branch)/len(segments) - np.sum(myelin_dimensions[:3])-np.sum(myelin_dimensions[4:])]+myelin_dimensions[4:])
			segmented_branch+=microstructure
			labellist+=labels
		
		del(myelin_dimensions)
		return(segmented_branch,labellist)
	
	def myelinate_branches(self,arbor,arbor_labels,mbranches,myelin_geometry=[1,1,3,8,3,1]):
		
		"""
		mbranches - branch indices in arbor to be myelinated
		myelin_dimensions - node length, paranode1 length, paranode2 length, internode (between paranodes) length
		return(arbor,labels)
		
		"""
		
		for branch in arbor.keys():
			if branch in mbranches:
				arbor[branch],arbor_labels[branch] = self.myelinate_branch(arbor[branch],myelin_geometry)
		
		return(arbor,arbor_labels)
	
	def insert_midpoint(self,a,b):
		midpoint = [item/2.0 for item in [a[0]+b[0],a[1]+b[1],a[2]+b[2],a[3]+b[3]]]
		return([list(a),midpoint,list(b)])
	
	def segment_and_label_branch(self,branch,label):
		if len(branch) < 3:
			newbranch = [self.insert_midpoint(branch[0],branch[-1])]
			return(newbranch,[label for i in range(len(newbranch))])
		
		else:
			return([self.insert_midpoint(branch[i],branch[i+1]) for i in range(len(branch[:-1]))],[label for i in range(len(branch[:-1]))])
	
	def bouton_branch(self,branch,bouton_dimensions):
		"""
		This function finds optimal microstructure dimensions to fit within a user specified branch, 
		redistributes points through the branch corresponding to the breakpoitns of these section,
		and returns the end points of each of these new structures.
		
		branch - list of points comprising a branch
		bouton_dimensions - descriptions of interbouton and bouton lengths (microns)
		returns - branch where list members are end points of each new section
		
		"""
		try:
			segments = self.find_optimal_segmentation_length(self.branchLength(branch),seg_range=np.sum(bouton_dimensions))
		except:
			return(self.segment_and_label_branch(branch,'interbouton'))
		
		if segments == []:
			return(self.segment_and_label_branch(branch,'interbouton'))
		
		try:
			splined_branch = self.spline_branch(branch,interval=segments[0])
		except:
			return(self.segment_and_label_branch(branch,'interbouton'))
		
		
		segmented_branch = []
		labellist = []
		for s,segment in enumerate(splined_branch[:-1]):
			microstructure,labels = self.divide_microstructure_unit(splined_branch[s],splined_branch[s+1],[self.branchLength(splined_branch)/len(segments) - bouton_dimensions[1],bouton_dimensions[1]])
			segmented_branch+=microstructure
			labellist+=labels
		
		del(bouton_dimensions)
		return(segmented_branch,labellist)
	
	
	def bouton_branches(self,arbor,arbor_labels,bbranches,bouton_geometry=[28,4]):
		"""
		
		bbranches - branch indices in arbor to be myelinated
		bouton_dimensions - (interbouton length upper and lower bounds), bouton length
		return(arbor,labels)
		
		"""
		
		for branch in arbor.keys():
			if branch in bbranches:
				arbor[branch],arbor_labels[branch] = self.bouton_branch(arbor[branch],bouton_geometry)
		
		return(arbor,arbor_labels)
	
	def calculate_start_middle_distances(self,point,newbranch):
		startdistances = []
		middledistances = []
		for section in newbranch:
			try:
				startdistances.append(self.eucdist3d(point,section[0]))
				middledistances.append(self.eucdist3d(point,section[1]))
			except:
				raise Exception(point,section,'couldnt calculate distance',newbranch)
		
		return(startdistances,middledistances)
	
	def add_back_branch_point(self,point,newbranch,labels):
		#find nearest section and nearest two points in section and insert b_point between them
		start,middle = self.calculate_start_middle_distances(point,newbranch)
		min_dist = np.max(middle)
		found = False
		for distance in zip(start,middle):
			if distance[1] < distance[0] and distance[0] < min_dist:
				min_dist = distance[1]
				found=True
		
		print(found,min_dist,'mindist',point,newbranch[middle.index(min_dist)])
		
#		if np.min(middle) > 40.0:
#			return(newbranch)
		
		split_branch = []
		split_labels = []
		if found:
			section_index = middle.index(min_dist)
		else:
			section_index = middle.index(np.min(middle))
		
		for s,section in enumerate(newbranch):
			if s == section_index:
				split_branch.append(self.insert_midpoint(section[0],point))
				split_branch.append(self.insert_midpoint(point,section[2]))
				split_labels.append(labels[s])
				split_labels.append(labels[s])
			else:
				split_branch.append(section)
				split_labels.append(labels[s])
		
#		if found:
#			newbranch[middle.index(min_dist)][1] = list(point)[:3] + [newbranch[middle.index(min_dist)][0][-1]]
#		else:
#			newbranch[middle.index(np.min(middle))][1] = list(point)[:3] + [newbranch[middle.index(min_dist)][0][-1]]
		
		
		
		return(newbranch,labels)
	
	def add_back_branch_points(self,originaltree,newtree,newlabels):
		b_points = [originaltree[b][0] for b in originaltree.keys()][1:]
		bbs = [b for b in originaltree.keys()][1:]
		for point in b_points:
			for branch in originaltree.keys():
				if branch == bbs[b_points.index(point)]:
					continue
				
				if point in originaltree[branch]:
					newtree[branch],newlabels[branch] =self.add_back_branch_point(point,newtree[branch],newlabels[branch])
		
		return(newtree,newlabels)
	
	def ensure_precision(self,arbor):
		for branch in arbor.keys():
			for s,section in enumerate(arbor[branch]):
				for p,point in enumerate(section):
					arbor[branch][s][p] = [round(item,4) for item in point]
		
		return(arbor)
	
	def add_microstructures_to_arbor(self,arbor,mbranches,bbranches):
		"""
		
		This function adds myelin and boutons to an axon arbor. 
		
		arbor - dictionary where keys are branch numbers and values are points comprising each branchLength
		mbranches - branch indices found in arbor where branches are expected to contain myelin_dimension
		bbranches - branch indices found in arbor where branches are expected to have en passant boutons
		
		returns: newly myelinated or boutoned arbor, and mirror dictionary with descriptive labels for all sections
		
		"""
		arbor_labels = dict(zip([key for key in arbor.keys()],[[] for key in arbor.keys()]))
		arbor,arbor_labels = self.myelinate_branches(arbor,arbor_labels,mbranches)
		arbor,arbor_labels = self.bouton_branches(arbor,arbor_labels,bbranches)
		arbor = self.ensure_precision(arbor)
		return(arbor,arbor_labels)
	
	def testfuncs(self):
		if self.eucdist3d((0,0,0),(1,0,0)) == 1.0:
			print('eucdist3d()--passed simple test')
		
		else:
			print('eucdist3d()--failed simple test')
		
		if self.branchLength([(0,0,0),(1,0,0)]) == 1.0:
			print('branchLength()--passed simple test')
		
		else:
			print('eucdist3d()--failed simple test')
		
		if self.spline_branch([(0,0,0),(1,0,0)],interval=0.5)==[(0,0,0),(0.5,0,0),(1,0,0)]:
			print('spline_branch()--passed simple test')
		
		else:
			print('eucdist3d()--failed simple test')
		
		if self.find_optimal_segmentation_length(112.215)[0] == 18.7025:
			print('find_optimal_segmentation_length()--passed simple test')
		
		else:
			print('find_optimal_segmentation_length()--failed simple test')
		
		sections,labels = self.divide_microstructure_unit([0,0,0],[10,0,0],[5,5])
		if sections == [[[0.0, 0.0, 0.0], [2.5, 0.0, 0.0], [5.0, 0.0, 0.0]], [[5.0, 0.0, 0.0], [7.5, 0.0, 0.0], [10, 0, 0]]]:
			print('divide_microstructure_unit() -boutons- --passed simple test')
		
		else:
			print('divide_microstructure_unit() -boutons- --failed simple test')
		
		sections,labels = self.divide_microstructure_unit([0,0,0],[19,0,0],[1,1,3,10,3,1])
		if sections == [[[0.0, 0.0, 0.0], [0.5, 0.0, 0.0], [1.0, 0.0, 0.0]], [[1.0, 0.0, 0.0], [1.5, 0.0, 0.0], [2.0, 0.0, 0.0]], [[2.0, 0.0, 0.0], [3.5, 0.0, 0.0], [5.0, 0.0, 0.0]], [[5.0, 0.0, 0.0], [10.0, 0.0, 0.0], [15.0, 0.0, 0.0]], [[15.0, 0.0, 0.0], [16.5, 0.0, 0.0], [18.0, 0.0, 0.0]], [[18.0, 0.0, 0.0], [18.5, 0.0, 0.0], [19, 0, 0]]]:
			print('divide_microstructure_unit() -myelin- --passed simple test')
		
		else:
			print('divide_microstructure_unit() -myelin- --failed simple test')
		
		branch,labels = self.myelinate_branch([[0,0,0],[100,0,0]])
		
		if branch == [[[0.0, 0.0, 0.0], [0.5, 0.0, 0.0], [1.0, 0.0, 0.0]], [[1.0, 0.0, 0.0], [1.5, 0.0, 0.0], [2.0, 0.0, 0.0]], [[2.0, 0.0, 0.0], [3.5, 0.0, 0.0], [5.0, 0.0, 0.0]], [[5.0, 0.0, 0.0], [17.166666666666664, 0.0, 0.0], [29.33333333333333, 0.0, 0.0]], [[29.33333333333333, 0.0, 0.0], [30.83333333333333, 0.0, 0.0], [32.33333333333333, 0.0, 0.0]], [[32.33333333333333, 0.0, 0.0], [32.83333333333333, 0.0, 0.0], [33.33333333333333, 0.0, 0.0]], [[33.33333333333333, 0.0, 0.0], [33.83333333333333, 0.0, 0.0], [34.33333333333333, 0.0, 0.0]], [[34.33333333333333, 0.0, 0.0], [34.83333333333333, 0.0, 0.0], [35.33333333333333, 0.0, 0.0]], [[35.33333333333333, 0.0, 0.0], [36.83333333333333, 0.0, 0.0], [38.33333333333333, 0.0, 0.0]], [[38.33333333333333, 0.0, 0.0], [50.49999999999999, 0.0, 0.0], [62.66666666666666, 0.0, 0.0]], [[62.66666666666666, 0.0, 0.0], [64.16666666666666, 0.0, 0.0], [65.66666666666666, 0.0, 0.0]], [[65.66666666666666, 0.0, 0.0], [66.16666666666666, 0.0, 0.0], [66.66666666666666, 0.0, 0.0]], [[66.66666666666666, 0.0, 0.0], [67.16666666666666, 0.0, 0.0], [67.66666666666666, 0.0, 0.0]], [[67.66666666666666, 0.0, 0.0], [68.16666666666666, 0.0, 0.0], [68.66666666666666, 0.0, 0.0]], [[68.66666666666666, 0.0, 0.0], [70.16666666666666, 0.0, 0.0], [71.66666666666666, 0.0, 0.0]], [[71.66666666666666, 0.0, 0.0], [83.83333333333333, 0.0, 0.0], [96.0, 0.0, 0.0]], [[96.0, 0.0, 0.0], [97.5, 0.0, 0.0], [99.0, 0.0, 0.0]], [[99.0, 0.0, 0.0], [99.5, 0.0, 0.0], [100.0, 0.0, 0.0]]]:
			print('myelinate_branch() --passed simple test')
		
		else:
			print('myelinate_branch() --failed simple test')
		
		branch,labels = self.bouton_branch([[0,0,0],[100,0,0]],bouton_dimensions=[(15,20),5])
		
		if branch == [[[0.0, 0.0, 0.0], [7.5, 0.0, 0.0], [15.0, 0.0, 0.0]], [[15.0, 0.0, 0.0], [17.5, 0.0, 0.0], [20.0, 0.0, 0.0]], [[20.0, 0.0, 0.0], [27.5, 0.0, 0.0], [35.0, 0.0, 0.0]], [[35.0, 0.0, 0.0], [37.5, 0.0, 0.0], [40.0, 0.0, 0.0]], [[40.0, 0.0, 0.0], [47.5, 0.0, 0.0], [55.00000000000001, 0.0, 0.0]], [[55.00000000000001, 0.0, 0.0], [57.50000000000001, 0.0, 0.0], [60.00000000000001, 0.0, 0.0]], [[60.00000000000001, 0.0, 0.0], [67.5, 0.0, 0.0], [75.0, 0.0, 0.0]], [[75.0, 0.0, 0.0], [77.5, 0.0, 0.0], [80.0, 0.0, 0.0]], [[80.0, 0.0, 0.0], [87.5, 0.0, 0.0], [95.0, 0.0, 0.0]], [[95.0, 0.0, 0.0], [97.5, 0.0, 0.0], [100.0, 0.0, 0.0]]]:
			print('bouton_branch() --passed simple test')
		
		else:
			print('bouton_branch() --failed simple test')





try:
	import networkx as nx
	import numpy as np

except:
	raise Exception('This class depends upon networkx, numpy. If you are unable to import this package, install it and try again.')


class GraphMethods():
	
	def __init__(self,faces,vertices,source_point,source_index,tri_edge_length_max=175):
		self.graph = self.build_graph(source_point,faces,vertices,tri_edge_length_max)
		source_index = [list(item) for item in list(vertices)].index(list(source_point))
		combos = [[source_index,i+1] for i in range(len(vertices[1:])) if i != source_index]
		self.node_dict,self.paths = self.find_most_likely_paths(source_index,combos)
	
	def euc_dist(self,A,B):
		return(((B[0]-A[0])**2 + (B[1]-A[1])**2 + (B[2]-A[2])**2)**0.5)
	
	def euc_dist_2d(self,A,B):
		return(((B[0]-A[0])**2 + (B[1]-A[1])**2)**0.5)
	
	def distance_filter_pass(self,A,B,threshold_dist):
		if self.euc_dist(A,B) > threshold_dist:
			return(False)
		else:
			return(True)
		
	def build_graph(self,source_point,faces,vertices,tri_edge_length_max):
		G = nx.Graph()
		for pyramid in faces:
			for pair in [[0,1],[0,2],[0,3],[1,2],[1,3],[2,3]]:
				try:
					if self.distance_filter_pass(vertices[pyramid[pair[0]]],vertices[pyramid[pair[1]]],tri_edge_length_max):
						try:
							G.add_edge(pyramid[pair[0]],pyramid[pair[1]],weight = self.euc_dist(vertices[pyramid[pair[0]]],vertices[pyramid[pair[1]]]))
						except:
							print(pyramid[pair[0]],pyramid[pair[1]],'coundt find it in vertices')
					
					###needs to be commented eventually because this is cheating to help arbitrary source points that may not be close to point cloud###
					elif list(source_point) in [list(vertices[pyramid[pair[0]]]),list(vertices[pyramid[pair[1]]])]:
						if self.distance_filter_pass(vertices[pyramid[pair[0]]],vertices[pyramid[pair[1]]],tri_edge_length_max+100):
							try:
								G.add_edge(pyramid[pair[0]],pyramid[pair[1]],weight = self.euc_dist(vertices[pyramid[pair[0]]],vertices[pyramid[pair[1]]]))
							except:
								print(pyramid[pair[0]],pyramid[pair[1]],'coundt find it in vertices')
				except:
					print(pyramid,list(faces).index(pyramid),'couldnt find it in vertices')
		
		return(G)
	
	def list_sort_by_size(self,LOLs):
		return(list(reversed(sorted(LOLs,key=len))))
	
	def less_than_count(self,incomplete_paths,p):
		return(p-len([l for l in incomplete_paths if l<p]))
	
	def find_longest_inclusive_paths(self,paths):
		paths = self.list_sort_by_size(paths)
		node_dict = {}
		for p,path in enumerate(paths):
			for node in path:
				if node not in node_dict.keys():
					node_dict[node] = p
		
		complete_paths = node_dict.values()
		incomplete_paths = [p for p,path in enumerate(paths) if p not in complete_paths]
		return(dict(zip(node_dict.values(),[self.less_than_count(incomplete_paths,p) for p in complete_paths])),[path for p,path in enumerate(paths) if p in complete_paths])
	
	def find_most_likely_paths(self,source_index,combos):
		paths = []
		for c,combo in enumerate(combos):
			try:
				paths.append(nx.dijkstra_path(self.graph,source_index,combo[1]))
			except:
				pass
		
		return(self.find_longest_inclusive_paths(paths))




import numpy as np
import math
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import random
from scipy.interpolate import interp1d
import scipy
import time
from roots.graphmethods import GraphMethods
from roots.kmd import KMDD
#import seaborn as sns
#plt.style.use('seaborn-deep')
#import matplotlib
#matplotlib.rcParams.update({'font.size': 32})
import os
random.seed("clayton\'s warhorse")
import traceback
from itertools import chain
#from colormap import Colormap
#c = Colormap()
#mycmap = c.cmap_linear('yellow','orange','red')


class Analyze_roots():
	def __init__(self,roots):
		self.roots = roots
		self.math = Roots_math()
		self.centroid = self.find_centroid()
		self.xyz_dist()
		self.internodedist_dist()
		self.branchnodes = [self.roots[node][0] for node in self.roots.keys()]
		self.seconds = [self.roots[node][1] for node in self.roots.keys()]
		self.thetas = self.bifurcation_thetas()
	
	def find_branch_parent(self,firstpoint,ownbnum):
		for brnum in self.roots.keys():
			if brnum != ownbnum and firstpoint in self.roots[brnum]:
				try:
					return(self.roots[brnum][self.roots[brnum].index(firstpoint)+1]) 
				except:
					print('parent not found',end="\r", flush=True)
					return(None)

	def find_branch_angle(self,firstpoint,firstpointsecond,secondsecond):	
		pointpresent = np.array(firstpoint)
		pointnext = np.array(firstpointsecond)
		pointlast = np.array(secondsecond)
		return(np.dot(pointlast-pointpresent,pointnext-pointpresent)/(self.math.eucdist3d(pointpresent,pointnext)*self.math.eucdist3d(pointpresent,pointlast)))
	
	
	def find_centroid(self):
		x,y,z = 0,0,0
		for i,key in enumerate(self.roots.keys()):
			for point in self.roots[key]:
				x+=point[0]
				y+=point[1]
				z+=point[2]
		return([x/float(i),y/float(i),z/float(i)])
	
	def xyz_dist(self):
		self.xdist,self.ydist,self.zdist = [],[],[]
		for i,key in enumerate(self.roots.keys()):
			for point in self.roots[key]:
				self.xdist.append(self.centroid[0] - point[0])
				self.ydist.append(self.centroid[1] - point[1])
				self.zdist.append(self.centroid[2] - point[2])
		ax = sns.distplot(np.array(self.xdist),kde=True, rug=True,label='X KDE', hist=False,kde_kws={"shade": True})
		ax = sns.distplot(np.array(self.ydist),kde=True, rug=True,label='Y KDE',hist=False,kde_kws={"shade": True})
		ax = sns.distplot(np.array(self.zdist),kde=True, rug=True, label='Z KDE',hist=False,kde_kws={"shade": True})
		plt.legend()
		plt.title('KDEs of Spatial Range of Arbor Relative to Centroid')
		plt.show()
	
	def bifurcation_thetas(self):
		thetas = []
		for n,node in enumerate(self.branchnodes):
			try:
				first,second,secondsecond = node,self.roots[n][1],self.find_branch_parent(node,n)
				#print(first,second,secondsecond)
				thetas.append(self.find_branch_angle(node,self.roots[n][1],self.find_branch_parent(node,n)))
			except:
				print('theta not found')
		ax = sns.distplot(np.abs(np.array(thetas)),kde=True, rug=True,label='Bifurcation Angles KDE', hist=False,kde_kws={"shade": True})
		plt.legend()
		plt.xlabel('Angle (Radians)')
		plt.show()
	
	def internodedist_dist(self):
		dists = []
		for i,key in enumerate(self.roots.keys()):
			for j,point in enumerate(self.roots[key][1:]):
				dists.append(self.math.eucdist3d(point,self.roots[key][j-1]))
		ax = sns.distplot(np.array(dists),label='Internode Distance', kde=True, rug=True,hist=False,kde_kws={"shade": True})
		plt.legend()
		plt.xlabel('um')
		plt.show()


class Roots_math():
	def __init__(self,isKMDD=False,KMDDproperties={},preconditionedKMDD=None):
		self.isKMDD = isKMDD
		self.KMDDproperties = KMDDproperties
		if self.isKMDD:
			try:
				cluster_reduce = self.KMDDproperties['cluster_reduce']
				tri_edge_length_max = self.KMDDproperties['tri_edge_length_max']
				openpoints = self.KMDDproperties['open_points']
				source = self.KMDDproperties['source']
			except:
				raise Exception("Error with KMDDproperties argument dictionary. Must include KMDDproperties values for keys: 'cluster_reduce','tri_edge_length_max','open_points','source'.")
			
			#find spatial clusters within cloud of synaptic targets in order to simplify arbor enough to find triangulation and paths through arbor field.
			#find delaunay triangulation through kmeans clustered spatial network of cloud of synaptic targets
			self.kmdd = KMDD(source,openpoints,cluster_reduce,tri_edge_length_max)
			
			#Dijkstras algorithm to find shortest paths through network of pairs of points in delaunay triangulated kmeans clustered cloud of synaptic targets.
			self.source_index = [list(item) for item in list(self.kmdd.vertices)].index(list(source))
			self.nk = GraphMethods(self.kmdd.faces,self.kmdd.vertices,source,self.source_index,tri_edge_length_max)
			if self.nk.node_dict == {}:
				raise Exception('Likely path sort failed. Try a longer tri_edge_length_max in KMDDproperties args.')
			
			print('got graph')
			try:
				assert self.KMDDproperties['plot']
				self.kmdd.plot_clusters_and_points(self.kmdd.vertices,openpoints)
				self.kmdd.mayavi_plot_delaunay_tri(source,tri_edge_length_max)
			except:
				pass
	
	def euc_tree(self,allpoints,point):
		return([self.eucdist3d(thisp,point) for thisp in allpoints])
	
	def euc_tree_length(self,arbor):
		len = 0
		for branch in arbor.keys():
			branchpoints = arbor[branch]
			for p,point in enumerate(branchpoints[1:]):
				try:
					len+=self.eucdist3d(branchpoints[p-1],branchpoints[p])
				
				except:
					print(traceback.format_exc())
		
		return(len)
	
	def eucdist3d(self,point1,point2):
		return(((point2[0]-point1[0])**2 + (point2[1]-point1[1])**2 + (point2[2]-point1[2])**2)**0.5)

	def angle_tree(self, allpoints, point, prevpoint):
		return([np.pi - self.check_angle(point, thisp, prevpoint) for thisp in allpoints])

	def choose_nearest(self, point, allpoints):
		""" chooses closest point from a set of given points and returns the index of that point in the array and the actual point itself """
		dists = dict(zip(self.euc_tree(allpoints,point),allpoints))
		if len(dists.keys()) > 1:
			nearest = dists[min(dists.keys())]
			nearestindex = allpoints.index(nearest)
		else:
			return(point, 0)
		
		return(nearest, nearestindex)

	def choose_nearest_bif(self, point, allpoints):
		dists = dict(zip(self.euc_tree(allpoints,point),allpoints))
		nearest = dists[min(dists.keys())]
		nearestindex = allpoints.index(nearest)
		return(nearest, nearestindex)
	
	def find_point_on_arbor(self, arbor, point, branch_counter):
		""" Finds branch_counter from last key and index of point on arbor and returns that key and index """
		
		counter = branch_counter
		for key in reversed(arbor.keys()):
			for index, this_point in enumerate(arbor[key]):
				if arbor[key][index] == point:
					if counter == 0:
						return(key, index)
					counter = counter - 1
		
	
	def lookup_nearest_likely_path(self,point_of_int):
		""" Returns a point a little over the specified path length away from the branch end """
		cluster_distances = self.euc_tree(self.kmdd.vertices,point_of_int)
		sorted_distances = list(sorted(cluster_distances))
		for increasing_distance in sorted_distances:
			nearest_cluster_node = cluster_distances.index(increasing_distance)
			for key in self.nk.node_dict.keys():
				if nearest_cluster_node in self.nk.paths[self.nk.node_dict[key]]:
					nearest_path_index = self.nk.node_dict[key]
					return(nearest_cluster_node,nearest_path_index)
		
		print('No path found for this cluster point',self.kmdd.vertices[nearest_cluster_node],end="\r", flush=True)
		#print('last path considered',self.nk.paths[self.nk.node_dict[key]])
		try:
			nearest_path_index = self.nk.node_dict[key]
		except:
			#nearest_path_index = self.nk.node_dict[list(self.nk.node_dict.keys())[0]]
			nearest_path_index = 0
		#nearest_path_index = self.node_dict[nearest_cluster_node]
		return(nearest_cluster_node,nearest_path_index)
	
	def find_diff_apply_to_other(self,A,B,C):
		diff = [B[0]-A[0],B[1]-A[1],B[2]-A[2]]
		return([C[0]-diff[0],C[1]-diff[1],C[2]-diff[2]])
	
	def find_dynamic_source_point_from_likely_path(self,point_of_int):
		nearest_cluster_node,nearest_path_index = self.lookup_nearest_likely_path(point_of_int)
		nearest_node_index_in_path = self.nk.paths[nearest_path_index].index(nearest_cluster_node)
		if nearest_node_index_in_path == len(self.nk.paths[nearest_path_index])-1:
			nearest_node_index_in_path-=1
		return(self.find_diff_apply_to_other(self.kmdd.vertices[self.nk.paths[nearest_path_index][nearest_node_index_in_path]],self.kmdd.vertices[self.nk.paths[nearest_path_index][nearest_node_index_in_path+1]],point_of_int))
	
	
	def find_relative_source(self, arbor, point_of_int, rel_source_dist):
		if self.isKMDD:
			try:
				return(self.find_dynamic_source_point_from_likely_path(point_of_int))
			
			except:
				print(traceback.format_exc())
				print('\n\nContinuing as if not KMDD directed',end="\r", flush=True)
				return(arbor[0][0])
		
		else:
			return(arbor[0][0])

	
	def branch_extension_check(self, firstextend, arborpoint, prevpoint, openpoints, distthresh, angthresh, rel_source_dist, arbor):
		""" Checks whether the branch can be extended; returns true or false and the point that it should be extended with """
		rel_source = self.find_relative_source(arbor, arborpoint, rel_source_dist)
		refdist = self.eucdist3d(arborpoint, rel_source) #distance from arbor point to source	
		alldists = self.euc_tree(openpoints, rel_source) #array of distances from source to each open point
		
		filteredpoints = [openpoints[i] for i, thisdist in enumerate(alldists) if thisdist >= refdist]
		dists = dict(zip(self.euc_tree(filteredpoints, arborpoint), filteredpoints)) #dictionary of distances between branch tip and set of open points paired with open points
		
		if firstextend:	
			if min(dists.keys()) > distthresh:
				return(False, dists[min(dists.keys())])
			
			else:
				return(True, dists[min(dists.keys())])
		
		else: #finds the closest point to branch tip, to extend from
			bad_ang = True
			good_dist = True
			nextpoint = [0,0,0]
			
			while (bad_ang and good_dist and len(dists.keys()) > 0):	
				bad_ang = np.pi - self.check_angle(arborpoint, dists[min(dists.keys())], prevpoint) > angthresh
				good_dist = min(dists.keys()) <= distthresh
				nextpoint = dists.pop(min(dists.keys()))
			
			return(not(bad_ang) and good_dist > 0, nextpoint)
	
	def return_closed_points(self,arbor,source): 
		closed = [item[1:-1] for item in list(arbor.values())]
		closed = [source]+list(chain.from_iterable(closed))
		return(closed)
	
	
	def choose_first_valid_bifurcation(self, arbor, openpoint, source, angthresh, distthresh, rel_source_dist):
		openpt_rel_source = self.find_relative_source(arbor,openpoint,rel_source_dist)
		openpt_rel_source_dist = self.eucdist3d(openpoint,openpt_rel_source)
		closed = self.return_closed_points(arbor,source)
		for c_point in closed[::-1]:
			if self.eucdist3d(c_point,openpt_rel_source) < openpt_rel_source_dist:
				filteredpoint = self.filter_angle(c_point,arbor,openpoint,angthresh)
				if filteredpoint == []:
					continue
				
				else:
					print(filteredpoint)
					return(True,filteredpoint)
	
		return(False,[0,0,0])
	
	
	
	def check_angle(self, pointpresent, pointnext, pointlast):
		pointpresent = np.array(pointpresent)
		pointnext = np.array(pointnext)
		pointlast = np.array(pointlast)
		return(np.abs(math.acos(np.dot(pointlast-pointpresent,pointnext-pointpresent)/(self.eucdist3d(pointpresent,pointnext)*self.eucdist3d(pointpresent,pointlast)))))

	def filter_angles(self, filteredpoints, arbor, openpoint, angthresh):
		morefilteredpoints = [] #array for output of function
		for fpoint in filteredpoints: #goes through each point in array, checks whether it is below angle threshold, and if so, appends point to morefilteredpoints
			goodangle = True
			for key in arbor.keys():
				if fpoint in arbor[key]:
					pointfoundindex = arbor[key].index(fpoint) #finds index of the point on the arbor in that key
					if self.check_angle(fpoint, arbor[key][pointfoundindex+1], openpoint) > angthresh: #checks whether angle meets threshold
						goodangle = False
						break
			
			if goodangle:
				morefilteredpoints.append(fpoint)
		
		return(morefilteredpoints)
	
	def filter_angle(self, filteredpoint, arbor, openpoint, angthresh):
		for key in arbor.keys():
			if filteredpoint in arbor[key]:
				pointfoundindex = arbor[key].index(filteredpoint) #finds index of the point on the arbor in that key
				if self.check_angle(arbor[key][pointfoundindex-1], filteredpoint, openpoint) > angthresh: #checks whether angle meets threshold
					return([])
				
				else:
					return(filteredpoint)
		
		return([])


class Roots():
	
	"""
	source - point from which the tree begins (not a member of open 'points' which is a separate argument)
	points - open points through which the tree must grow (does not include 'source')
	s_ang - branch extension angle threshold
	s_dist - branch extension distance threshold
	b_ang - bifurcation angle threshold
	b_dist - bifurcation distance threshold
	rel_source_dist - (disabled for KMDD mode growing) minimum distance down the active branch on which to find a bifurcation point
	KMDDproperties - properties of k-means, delaunay triangulation for 'likely path' sorting for dynamic source updating fibers provided as a dictionary
		KMDDproperties:
				cluster_reduce - proportion of reduction for k-means clustering (i.e. if cluster_reduce == 0.25 then the number of clusters will equal 25% of the number of open points)
				tri_edge_length_max - the meshing distance threshold below which delaunay triangulated edges will be returned (i.e. triangulated edges longer than tri_edge_length_max will be removed from the triangulation)
				openpoints - open points to be clustered and meshed (same as ROOTS argument 'points')
				source - source point from which a tree will be grown (same as ROOTS argument 'source')
	"""
	def __init__(self, source, points,total_length=np.inf, s_ang=np.pi, s_dist=100, b_ang=np.pi, b_dist=100, bnum_limit=np.inf,rel_source_dist=100,KMDDproperties={}):
		self.KMDDproperties = KMDDproperties
		self.source = source
		self.openpoints = points
		self.tot_len_threshold = total_length
		self.bnum_limit = bnum_limit
		self.s_ang = s_ang
		self.s_dist = s_dist
		self.b_ang = b_ang
		self.b_dist = b_dist
		self.rel_source_dist = rel_source_dist
		self.allpoints = [self.source]+self.openpoints
		self.arbor = {0:[self.source]}
		self.stamps = []
		self.labels = []
		if self.KMDDproperties == {}:
			self.math = Roots_math(False,self.KMDDproperties,None)
			self.sort_points()
		
		else:
			self.math = Roots_math(True,self.KMDDproperties,None)
			self.sort_points()
		
		#self.plotroots = Roots_plot()
	
	
	def stamp(self,label):
		self.stamps.append(time.time())
		self.labels.append(label)
		
	
	def sort_points(self):
		self.stamp('sort points')
		tpoints = list(zip(self.openpoints,self.math.euc_tree(self.openpoints, self.source)))
		tpoints.sort(key= lambda x: x[1])
		self.openpoints = [tpoint[0] for tpoint in tpoints]
		self.stamp('sort points')
	
	def clean_points(self,point):
		pointindex = self.openpoints.index(point)
		self.openpoints = self.openpoints[:pointindex] + self.openpoints[pointindex+1:]
	
	def bifurcate(self,branchpoint, newpoint):
		self.arbor[len(self.arbor.keys())] = [branchpoint, newpoint] #appends new key at end of arbor as a new branch. This is an array of two points, the point previously on the arbor and a new point branching off from this point
	
	def is_acceptable_total_length(self):
		if self.calculate_total_arbor_length() > self.tot_len_threshold:
			return(False)
		
		else:
			return(True)
	
	def is_acceptable_number_of_branches(self):
		if len(self.arbor.keys()) > self.bnum_limit:
			return(False)
		
		else:
			return(True)
	
	
	def calculate_total_arbor_length(self):
		return(self.math.euc_tree_length(self.arbor))
	
	def add_to_arbor(self,newpoint):
		arborindex = max(list(self.arbor.keys()))
		self.arbor[arborindex].append(newpoint)
	
	
	def prims(self):
		arborindex = max(list(self.arbor.keys())) #last key in arbor
		number_of_points = 0
		stillextending = True
		
		while stillextending:
			if len(self.arbor[arborindex]) == 1: #initial branch extension from soma
				#goodextend, nearest = self.math.branch_extension_check(True, self.arbor[arborindex][-1], self.arbor[arborindex][-1], self.openpoints, self.s_dist, self.s_ang, self.source)
				goodextend, nearest = self.math.branch_extension_check(True, self.arbor[arborindex][-1], self.arbor[arborindex][-1], self.openpoints, self.s_dist, self.s_ang, self.rel_source_dist, self.arbor)
			
			else:
				#goodextend, nearest = self.math.branch_extension_check(False, self.arbor[arborindex][-1], self.arbor[arborindex][-2], self.openpoints, self.s_dist, self.s_ang, self.source)
				goodextend, nearest = self.math.branch_extension_check(False, self.arbor[arborindex][-1], self.arbor[arborindex][-2], self.openpoints, self.s_dist, self.s_ang, self.rel_source_dist, self.arbor)
			
			if goodextend:
				self.add_to_arbor(nearest) #adds point to end of last key of the arbor
				self.clean_points(nearest) #gets rid of added point from the set of open points
				number_of_points = number_of_points + 1
				stillextending = self.is_acceptable_total_length()
				continue
			
			stillextending = False
		
		return(number_of_points)
	
	def find_branch_point(self):
		goodbranch = False
		for newpoint in self.openpoints: #goes through each open point
			goodbranch, nearest = self.math.choose_first_valid_bifurcation(self.arbor, newpoint, self.source, self.b_ang, self.b_dist, self.rel_source_dist)
			#print(goodbranch,nearest)
			if goodbranch:	
				self.bifurcate(nearest, newpoint)
				self.clean_points(newpoint)
				goodbranch = True
				#print('bifurcated')
				break
		
		return(goodbranch)
	
	def grow(self):
		goodbranch = True
		number_of_points = 0
		self.stamp('grow')
		while len(self.openpoints) > 0 and goodbranch: #grows arbor using prims and find_branch_point methods, till there are no more open points or goodbranch is false
			self.stamp('prims')
			number_of_points = number_of_points + self.prims()
			if not self.is_acceptable_total_length():
				return(number_of_points)
			
			if not self.is_acceptable_number_of_branches():
				return(number_of_points)
			
			self.stamp('prims')
			self.stamp('bifurcate')
			goodbranch = self.find_branch_point()
			self.stamp('bifurcate')
			print(str(round(100.0*float(number_of_points)/float(len(self.allpoints)),2))+'%'+' connected',end="\r", flush=True)
		self.stamp('grow')
		return(number_of_points)





try:
	from mayavi import mlab
	import numpy as np
except:
	print('Some elements of this package depend on mayavi2,numpy. Install mayavi to ensure mlab can be successfully imported and try again.')

try:
	import matplotlib.pyplot as plt
	from mpl_toolkits.mplot3d import Axes3D
except:
	print('Some elements of this package depend upon matplotlib and mpl_toolkits.mplot3d. If you are unable to import these, install them and try again.')

class KMDD():
	def __init__(self,source_point,points_outer,cluster_reduce=0.15,tri_edge_length_max=175):
		self.source_point = source_point
		self.points_outer = points_outer
		self.cluster_reduce = cluster_reduce
		self.tri_edge_length_max = tri_edge_length_max
		self.vertices,self.faces = self.return_meshed_point_cloud(source_point,points_outer,cluster_reduce)
#		self.plot_clusters_and_points(self.vertices,points_outer)
		self.mayavi_plot_delaunay_tri(source_point,tri_edge_length_max)
	
	def plot_clusters_and_points(self,clusters,points):
		fig = plt.figure()
		ax = Axes3D(fig)
		ax.scatter([point[0] for point in clusters],[point[1] for point in clusters],[point[2] for point in clusters],'.r',label='All Points',s=3.5)
		ax.scatter([point[0] for point in points[::5]],[point[1] for point in points[::5]],[point[2] for point in points[::5]],'.k',label='All Points',s=2)
		plt.show()

	def euc_dist(self,A,B):
		return(((B[0]-A[0])**2 + (B[1]-A[1])**2 + (B[2]-A[2])**2)**0.5)

	def euc_dist_2d(self,A,B):
		return(((B[0]-A[0])**2 + (B[1]-A[1])**2)**0.5)

	def distance_filter_pass(self,A,B,threshold_dist):
		if self.euc_dist(A,B) > threshold_dist:
			return(False)
		else:
			return(True)

	def return_meshed_point_cloud(self,sourcepoint, points,simplify=0.1):
		#from sklearn.cluster import KMeans as kmeans
		from sklearn.cluster import MiniBatchKMeans as mbkmeans
		from scipy.spatial import Delaunay as delaunay
		nclusters = int(len(points)*simplify)
		#km = kmeans(n_clusters=nclusters, init='k-means++', n_init=10, max_iter=300, tol=0.0001, precompute_distances='auto', verbose=0, random_state=None, copy_x=True, n_jobs=1, algorithm='auto').fit(points) 
		km = mbkmeans(n_clusters=nclusters, init='k-means++', n_init=10, max_iter=300, tol=0.001, verbose=0, random_state=None, init_size=3*nclusters).fit(points) 
		self.clusters = km.cluster_centers_
		clusters = [list(item) for item in list(km.cluster_centers_)]
		if list(sourcepoint) in clusters:
			print('sourcepoint already in clusters')
		else:
			clusters=[list(sourcepoint)]+clusters
		
		print('got '+str(len(clusters))+' clusters')
		tri_volume = delaunay(np.array(clusters), furthest_site=False, incremental=False, qhull_options=None)
		return(tri_volume.points,tri_volume.simplices)
	
	def mayavi_plot_delaunay_tri(self,source_point,dist_threshold):
		src = mlab.pipeline.scalar_scatter(np.array([pt[0] for pt in self.vertices]), np.array([pt[1] for pt in self.vertices]), np.array([pt[2] for pt in self.vertices]), np.array([1 for pt in self.vertices]))
		connections = []
		for pyramid in self.faces:
			for pair in [[0,1],[0,2],[0,3],[1,2],[1,3],[2,3]]:
				if self.distance_filter_pass(self.vertices[pyramid[pair[0]]],self.vertices[pyramid[pair[1]]],dist_threshold):
					connections.append(np.array([pyramid[pair[0]],pyramid[pair[1]]]))
					continue
				
				if list(source_point) in [list(self.vertices[pyramid[pair[0]]]),list(self.vertices[pyramid[pair[1]]])]:
					if self.distance_filter_pass(self.vertices[pyramid[pair[0]]],self.vertices[pyramid[pair[1]]],dist_threshold+100):
						connections.append(np.array([pyramid[pair[0]],pyramid[pair[1]]]))
		
		
		connections = np.vstack(tuple(connections))
		src.mlab_source.dataset.lines = connections
		src.update()
		lines = mlab.pipeline.stripper(src)
		mlab.pipeline.surface(lines, colormap='Accent', line_width=1, opacity=.4)
		mlab.show()



import os

class Root2Hoc():
	"""
	This class is designed to take an existing morphology and write the topology and morphology to a .hoc file that can be used to run NEURON simulations using existing models.
	
	```
	hoc_writer = Root2Hoc()
	hoc_writer.primarytest()
	arbor = hoc_writer.test_arbor()
	print(arbor)
	hoc_writer.arbor_to_nrn(arbor)
	```
	
	"""
	
	def __init__(self):
		pass
	
	def strip_diameters_for_comparison(self,li):
		if any(isinstance(i, list) for i in li):
			newli = []
			if any(isinstance(i, list) for i in li[0]):
				for section in li:
					newli.append([point[:3] for point in section])
				
				return(newli)
			
			else:
				return([point[:3] for point in li])
		
		else:
			return(li[:3])
	
	def lookup_section_by_point(self,arbor,point,home_branch):
		sec_index = 0
		for branch in arbor.keys():
			if branch == home_branch:
				continue
			for section in arbor[branch]:
				if self.strip_diameters_for_comparison(point) in self.strip_diameters_for_comparison(section):
					return(sec_index,self.strip_diameters_for_comparison(section).index(self.strip_diameters_for_comparison(point))/float(len(section)-1))
				
				else:
					sec_index+=1
		
		return(None,None)
	
	def count_sections(self,arbor):
		sec_count = 0
		for branch in arbor.keys():
			for section in arbor[branch]:
				sec_count+=1
		
		return(sec_count)
	
	def build_section_indices(self,arbor):
		"""
		
		builds a dictionary of indices that match the arbor sections and is used to label sections while building the connections/topology
		
		"""
		
		section_indices = {}
		sec_ind = 0
		for branch in arbor.keys():
			section_indices[branch] = []
			for section in arbor[branch]:
				section_indices[branch].append(sec_ind)
				sec_ind+=1
		
		return(section_indices)
	
	def build_topology_list(self,arbor):
		section_indices = self.build_section_indices(arbor)
		connections = []
		for branch in arbor.keys():
			if branch != 0:
				ind,position = self.lookup_section_by_point(arbor,arbor[branch][0][0],branch)
				if ind==None:
					raise(ValueError)
				
				connections.append(([ind,int(position)],[section_indices[branch][0],0]))
			
			for section in arbor[branch][:-1]:
				connections.append(([section_indices[branch][arbor[branch].index(section)],1],[section_indices[branch][arbor[branch].index(section)]+1,0]))
		
		return(connections)
	
	def arbor_to_nrn_topology(self,arbor):
		topo_list = self.build_topology_list(arbor)
		number_of_sections = self.count_sections(arbor)
		topo_func = '//Building axon topology\n\n'
		topo_func+= f'create sectionList[{number_of_sections}]'
		
		for connection in topo_list:
			topo_func+=f'\nconnect sectionList[{connection[1][0]}]({connection[1][1]}), sectionList[{connection[0][0]}]({connection[0][1]})'
		
		return(topo_func+'\n')
	
	def nrn_build_section(self,section_index,section_points):
		nrn_command=f'sectionList[{section_index}]'+'{\n'
		for point in section_points:
			if len(point) < 4:
				nrn_command+=f'\tpt3dadd({point[0]},{point[1]},{point[2]},1)\n'
			else:
				nrn_command+=f'\tpt3dadd({point[0]},{point[1]},{point[2]},{point[3]})\n'
		
		
		return(nrn_command+'}\n')
	
	def arbor_to_nrn_morphology(self,arbor):
		"""
		//Set up axon geometry
		
		
		//MAIN AXON
		node[0]{
		    pt3dadd(-10338,2909.5,-316.24,nodeD)
		    pt3dadd(-10338.3947,2910.3684,-315.94,nodeD)
		}
		"""
		morpho_func='\n//Set up axon geometry\n\n//MAIN AXON\n\n'
		section_index = 0
		for branch in arbor.keys():
			for section in arbor[branch]:
				morpho_func+=self.nrn_build_section(section_index,section)
				section_index+=1
		
		
		
		return(morpho_func)
	
	
	def arbor_to_nrn(self,arbor,labels,target=os.getcwd()+'/testmorphology.hoc'):
		topology_function = self.arbor_to_nrn_topology(arbor)
		morphology_function = self.arbor_to_nrn_morphology(arbor)
		with open(target,'w') as f:
			f.write(topology_function+'\n\n'+morphology_function+'\n\n')
		
		return(topology_function+'\n\n'+morphology_function+'\n\n')
	
	def test_arbor(self):
		arbor = {}
		arbor[0] = [[[0,0,0,1],[1,0,0,1],[2,0,0,1]],[[2,0,0,1],[3,0,0,1],[4,0,0,1]],[[4,0,0,1],[5,0,0,1],[6,0,0,1]]]
		arbor[1] = [[[2,0,0,1],[2,1,0,1],[2,2,0,1]],[[2,2,0,1],[2,3,0,1],[2,4,0,1]]]
		arbor[2] = [[[4,0,0,1],[4,0,1,1],[4,0,2,1]],[[4,0,2,1],[4,0,3,1],[4,0,4,1]]]
		
		
		return(arbor)
	
	def primarytest(self):
		arbor = self.test_arbor()
		if self.lookup_section_by_point(arbor,[3,0,0,1])[0] == 1:
			print('lookup_section_by_point() --passed')
		
		else:
			print('lookup_section_by_point() --failed')
		
		if self.count_sections(arbor) == 7:
			print('count_sections() --passed')
		
		else:
			print('count_sections() --failed')
		
		if self.build_topology_list(arbor) == [([0, 1], [1, 0]), ([1, 1], [2, 0]), ([0, 1], [3, 0]), ([3, 1], [4, 0]), ([1, 1], [5, 0]), ([5, 1], [6, 0])]:
			print('build_section_list() --passed\nbuild_section_indices() --passed')
		
		else:
			print('build_section_list() --failed\nbuild_section_indices() --failed')
		
		if self.arbor_to_nrn_topology(arbor) == '//Building axon topology\n\ncreate sectionList[7]\nconnect sectionList[0](1), sectionList[1](0)\nconnect sectionList[1](1), sectionList[2](0)\nconnect sectionList[0](1), sectionList[3](0)\nconnect sectionList[3](1), sectionList[4](0)\nconnect sectionList[1](1), sectionList[5](0)\nconnect sectionList[5](1), sectionList[6](0)\n':
			print('arbor_to_nrn_topology() --passed')
		
		else:
			print('arbor_to_nrn_topology() --failed')
		
		if self.nrn_build_section(0,arbor[0][0]) =='sectionList[0]{\n\tpt3dadd(0,0,0,1)\n\tpt3dadd(1,0,0,1)\n\tpt3dadd(2,0,0,1)\n}\n':
			print('nrn_build_section() --passed')
		
		else:
			print('nrn_build_section() --failed')
		
		if self.arbor_to_nrn_morphology(arbor) =='\n//Set up axon geometry\n\n//MAIN AXON\n\nsectionList[0]{\n\tpt3dadd(0,0,0,1)\n\tpt3dadd(1,0,0,1)\n\tpt3dadd(2,0,0,1)\n}\nsectionList[1]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(3,0,0,1)\n\tpt3dadd(4,0,0,1)\n}\nsectionList[2]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(5,0,0,1)\n\tpt3dadd(6,0,0,1)\n}\nsectionList[3]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(2,1,0,1)\n\tpt3dadd(2,2,0,1)\n}\nsectionList[4]{\n\tpt3dadd(2,2,0,1)\n\tpt3dadd(2,3,0,1)\n\tpt3dadd(2,4,0,1)\n}\nsectionList[5]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(4,0,1,1)\n\tpt3dadd(4,0,2,1)\n}\nsectionList[6]{\n\tpt3dadd(4,0,2,1)\n\tpt3dadd(4,0,3,1)\n\tpt3dadd(4,0,4,1)\n}\n':
			print('arbor_to_nrn_morphology() --passed')
		
		else:
			print('arbor_to_nrn_morphology() --failed')
		
		if self.arbor_to_nrn(arbor) == '//Building axon topology\n\ncreate sectionList[7]\nconnect sectionList[0](1), sectionList[1](0)\nconnect sectionList[1](1), sectionList[2](0)\nconnect sectionList[0](1), sectionList[3](0)\nconnect sectionList[3](1), sectionList[4](0)\nconnect sectionList[1](1), sectionList[5](0)\nconnect sectionList[5](1), sectionList[6](0)\n\n\n\n//Set up axon geometry\n\n//MAIN AXON\n\nsectionList[0]{\n\tpt3dadd(0,0,0,1)\n\tpt3dadd(1,0,0,1)\n\tpt3dadd(2,0,0,1)\n}\nsectionList[1]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(3,0,0,1)\n\tpt3dadd(4,0,0,1)\n}\nsectionList[2]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(5,0,0,1)\n\tpt3dadd(6,0,0,1)\n}\nsectionList[3]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(2,1,0,1)\n\tpt3dadd(2,2,0,1)\n}\nsectionList[4]{\n\tpt3dadd(2,2,0,1)\n\tpt3dadd(2,3,0,1)\n\tpt3dadd(2,4,0,1)\n}\nsectionList[5]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(4,0,1,1)\n\tpt3dadd(4,0,2,1)\n}\nsectionList[6]{\n\tpt3dadd(4,0,2,1)\n\tpt3dadd(4,0,3,1)\n\tpt3dadd(4,0,4,1)\n}\n\n\n':
			print('arbor_to_nrn() --passed')
		
		else:
			print('arbor_to_nrn() --failed')




class Root2Py():
	"""
	This class is designed to take an existing morphology and write the topology and morphology to a .py file that can be used to run NEURON simulations using existing models.
	
	```
	nrn_writer = Root2Py()
	nrn_writer.primarytest()
	arbor = nrn_writer.test_arbor()
	print(arbor)
	nrn.arbor_to_nrn(arbor)
	```
	
	"""
	
	def __init__(self):
		pass
	
	def strip_diameters_for_comparison(self,li):
		if any(isinstance(i, list) for i in li):
			newli = []
			if any(isinstance(i, list) for i in li[0]):
				for section in li:
					newli.append([point[:3] for point in section])
				return(newli)
			
			else:
				return([point[:3] for point in li])
		
		else:
			return(li[:3])
	
	def lookup_section_by_point(self,arbor,point,home_branch):
		sec_index = 0
		for branch in arbor.keys():
			if branch == home_branch:
				continue
			for section in arbor[branch]:
				if self.strip_diameters_for_comparison(point) in self.strip_diameters_for_comparison(section):
					return(sec_index,self.strip_diameters_for_comparison(section).index(self.strip_diameters_for_comparison(point))/float(len(section)-1))
				
				else:
					sec_index+=1
		
		return(None,None)
	
	def count_sections(self,arbor):
		sec_count = 0
		for branch in arbor.keys():
			for section in arbor[branch]:
				sec_count+=1
		
		return(sec_count)
	
	def build_section_indices(self,arbor):
		"""
		
		builds a dictionary of indices that match the arbor sections and is used to label sections while building the connections/topology
		
		"""
		
		section_indices = {}
		sec_ind = 0
		for branch in arbor.keys():
			section_indices[branch] = []
			for section in arbor[branch]:
				section_indices[branch].append(sec_ind)
				sec_ind+=1
		
		return(section_indices)
	
	def build_topology_list(self,arbor):
		section_indices = self.build_section_indices(arbor)
		connections = []
		for branch in arbor.keys():
			if branch != 0:
				ind,position = self.lookup_section_by_point(arbor,arbor[branch][0][0],branch)
				if ind==None:
					raise(ValueError)
				
				connections.append(([ind,int(position)],[section_indices[branch][0],0]))
			
			for section in arbor[branch][:-1]:
				connections.append(([section_indices[branch][arbor[branch].index(section)],1],[section_indices[branch][arbor[branch].index(section)]+1,0]))
		
		return(connections)
	
	def arbor_to_nrn_topology(self,arbor):
		topo_list = self.build_topology_list(arbor)
		number_of_sections = self.count_sections(arbor)
		topo_func = '#Building axon topology\n\nfrom neuron import h\n\n'
		topo_func+= f'sectionList = [h.Section() for i in range({number_of_sections})]'
		
		for connection in topo_list:
			topo_func+=f'\nsectionList[{connection[1][0]}].connect(sectionList[{connection[0][0]}]({connection[0][1]}),{connection[1][1]})'
		
		return(topo_func+'\n')
	
	def nrn_build_section(self,section_index,section_points):
		nrn_command='\n\n'
		for point in section_points:
			if len(point) < 4:
				nrn_command+=f'h.pt3dadd({point[0]},{point[1]},{point[2]},1,sec=sectionList[{section_index}])\n'
			else:
				nrn_command+=f'h.pt3dadd({point[0]},{point[1]},{point[2]},{point[3]},sec=sectionList[{section_index}])\n'
		
		
		return(nrn_command)
	
	def arbor_to_nrn_morphology(self,arbor):
		"""
		#Set up axon geometry
		
		
		#MAIN AXON
		node[0] = h.Section()
		h.pt3dadd(-10338,2909.5,-316.24,nodeD,sec=node[0])
		h.pt3dadd(-10338.3947,2910.3684,-315.94,nodeD,sec=node[0])
		
		"""
		nrn_morpho_func='\n#Set up axon geometry\n\n#MAIN AXON\n\n'
		section_index = 0
		for branch in arbor.keys():
			for section in arbor[branch]:
				nrn_morpho_func+=self.nrn_build_section(section_index,section)
				section_index+=1
		
		
		
		return(nrn_morpho_func)
	
	
	def arbor_to_nrn(self,arbor,labels,target=os.getcwd()+'/testmorphology.py'):
		topology_function = self.arbor_to_nrn_topology(arbor)
		morphology_function = self.arbor_to_nrn_morphology(arbor)
		with open(target,'w') as f:
			f.write(topology_function+'\n\n'+morphology_function+'\n\n')
		
		return(topology_function+'\n\n'+morphology_function+'\n\n')
	
	def test_arbor(self):
		arbor = {}
		arbor[0] = [[[0,0,0,1],[1,0,0,1],[2,0,0,1]],[[2,0,0,1],[3,0,0,1],[4,0,0,1]],[[4,0,0,1],[5,0,0,1],[6,0,0,1]]]
		arbor[1] = [[[2,0,0,1],[2,1,0,1],[2,2,0,1]],[[2,2,0,1],[2,3,0,1],[2,4,0,1]]]
		arbor[2] = [[[4,0,0,1],[4,0,1,1],[4,0,2,1]],[[4,0,2,1],[4,0,3,1],[4,0,4,1]]]
		
		
		return(arbor)
	
	def primarytest(self):
		arbor = self.test_arbor()
		if self.lookup_section_by_point(arbor,[3,0,0,1])[0] == 1:
			print('lookup_section_by_point() --passed')
		
		else:
			print('lookup_section_by_point() --failed')
		
		if self.count_sections(arbor) == 7:
			print('count_sections() --passed')
		
		else:
			print('count_sections() --failed')
		
		if self.build_topology_list(arbor) == [([0, 1], [1, 0]), ([1, 1], [2, 0]), ([0, 1], [3, 0]), ([3, 1], [4, 0]), ([1, 1], [5, 0]), ([5, 1], [6, 0])]:
			print('build_section_list() --passed\nbuild_section_indices() --passed')
		
		else:
			print('build_section_list() --failed\nbuild_section_indices() --failed')
		
		if self.arbor_to_nrn_topology(arbor) == '//Building axon topology\n\ncreate sectionList[7]\nconnect sectionList[0](1), sectionList[1](0)\nconnect sectionList[1](1), sectionList[2](0)\nconnect sectionList[0](1), sectionList[3](0)\nconnect sectionList[3](1), sectionList[4](0)\nconnect sectionList[1](1), sectionList[5](0)\nconnect sectionList[5](1), sectionList[6](0)\n':
			print('arbor_to_nrn_topology() --passed')
		
		else:
			print('arbor_to_nrn_topology() --failed')
		
		if self.nrn_build_section(0,arbor[0][0]) =='sectionList[0]{\n\tpt3dadd(0,0,0,1)\n\tpt3dadd(1,0,0,1)\n\tpt3dadd(2,0,0,1)\n}\n':
			print('nrn_build_section() --passed')
		
		else:
			print('nrn_build_section() --failed')
		
		if self.arbor_to_nrn_morphology(arbor) =='\n//Set up axon geometry\n\n//MAIN AXON\n\nsectionList[0]{\n\tpt3dadd(0,0,0,1)\n\tpt3dadd(1,0,0,1)\n\tpt3dadd(2,0,0,1)\n}\nsectionList[1]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(3,0,0,1)\n\tpt3dadd(4,0,0,1)\n}\nsectionList[2]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(5,0,0,1)\n\tpt3dadd(6,0,0,1)\n}\nsectionList[3]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(2,1,0,1)\n\tpt3dadd(2,2,0,1)\n}\nsectionList[4]{\n\tpt3dadd(2,2,0,1)\n\tpt3dadd(2,3,0,1)\n\tpt3dadd(2,4,0,1)\n}\nsectionList[5]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(4,0,1,1)\n\tpt3dadd(4,0,2,1)\n}\nsectionList[6]{\n\tpt3dadd(4,0,2,1)\n\tpt3dadd(4,0,3,1)\n\tpt3dadd(4,0,4,1)\n}\n':
			print('arbor_to_nrn_morphology() --passed')
		
		else:
			print('arbor_to_nrn_morphology() --failed')
		
		if self.arbor_to_nrn(arbor) == '//Building axon topology\n\ncreate sectionList[7]\nconnect sectionList[0](1), sectionList[1](0)\nconnect sectionList[1](1), sectionList[2](0)\nconnect sectionList[0](1), sectionList[3](0)\nconnect sectionList[3](1), sectionList[4](0)\nconnect sectionList[1](1), sectionList[5](0)\nconnect sectionList[5](1), sectionList[6](0)\n\n\n\n//Set up axon geometry\n\n//MAIN AXON\n\nsectionList[0]{\n\tpt3dadd(0,0,0,1)\n\tpt3dadd(1,0,0,1)\n\tpt3dadd(2,0,0,1)\n}\nsectionList[1]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(3,0,0,1)\n\tpt3dadd(4,0,0,1)\n}\nsectionList[2]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(5,0,0,1)\n\tpt3dadd(6,0,0,1)\n}\nsectionList[3]{\n\tpt3dadd(2,0,0,1)\n\tpt3dadd(2,1,0,1)\n\tpt3dadd(2,2,0,1)\n}\nsectionList[4]{\n\tpt3dadd(2,2,0,1)\n\tpt3dadd(2,3,0,1)\n\tpt3dadd(2,4,0,1)\n}\nsectionList[5]{\n\tpt3dadd(4,0,0,1)\n\tpt3dadd(4,0,1,1)\n\tpt3dadd(4,0,2,1)\n}\nsectionList[6]{\n\tpt3dadd(4,0,2,1)\n\tpt3dadd(4,0,3,1)\n\tpt3dadd(4,0,4,1)\n}\n\n\n':
			print('arbor_to_nrn() --passed')
		
		else:
			print('arbor_to_nrn() --failed')




import setuptools
from distutils.core import setup

DESCRIPTION = "Grows artificial axon arbors based on morphometrics extracted from full axon reconstructions."
LONG_DESCRIPTION = "ROOTS: Ruled-Optimum Ordered Tree System. ROOTS is a package designed to take experimentally determined morphometric data and return an artificial neuronal arbor which satisfies the user-defined parameters. Originally this package was written to aid the construction of biologically realistic neural models for use in computational studies of extracellular electrical stimulation but it can be used for diverse neural modeling problems."
NAME = "Roots"
AUTHOR = "Clayton Bingham"
AUTHOR_EMAIL = "clayton.bingham@gmail.com"
MAINTAINER = "Clayton Bingham"
MAINTAINER_EMAIL = "clayton.bingham@gmail.com"
LICENSE = 'MIT'
REQUIREMENTS = [
		'numpy',
		'matplotlib',
		'networkx',
		'mayavi',
		'mpl_toolkits',
		'scipy',
		'scikit-learn'
		]
VERSION = '0.0.1'

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      download_url="https://github.com/bingsome/Roots.git",
      license=LICENSE,
      packages=['roots'],
      install_requires=REQUIREMENTS,
      package_data={}
     )



import numpy as np
from mayavi import mlab
import os
from roots.swcToolkit import swcToolkit


class swcVisualizer():
	"""
	
	mfile = 'fileonpath.swc'
	visualizer = swcVisualizer()
	visualizer.mplot_mfile(mfile)
	
	
	"""
	
	def __init__(self):
		self.swcTool = swcToolkit()
	
	def create_cylinders(self,coords,diams,data,num_pts):
		x = []
		y = []
		z = []
		connections = []
		D = []
		offset = 0
		for kk in range(len(coords)):
			# Define points
			C1 = np.array(coords[kk][0])
			C2 = np.array(coords[kk][1])
			
			# Define normal plane
			p = C1-C2
			d = np.dot(p,C1)
			
			# Get normal vectors on plane
			z_idx = np.arange(3)[p==0]
			nz_idx = np.arange(3)[p!=0]
			if len(nz_idx) == 3:
				x1 = 1.
				y1 = 1.
				z1 = (d-(np.dot(p[:2],[x1,y1])))/p[2]
				a = np.array([x1,y1,z1])
			elif len(nz_idx) == 2:
				a = np.zeros(3)
				a[z_idx] = 1.
				a[nz_idx[0]] = 1.
				a[nz_idx[1]] = (d-p[nz_idx[0]])/p[nz_idx[1]]	
			else:
				a = np.zeros(3)
				a[z_idx] = 1.
				a[nz_idx] = d/p[nz_idx]
			a = a-C1
			if len(p[p!=0]) == 3:
				x2 = 1.
				y2 = (a[2]*p[0]/p[2] - a[0]) / (a[1] - a[2]*p[1]/p[2])
				z2 = -(p[1]*y2+p[0])/p[2]
				b = np.array([x2,y2,z2])
			elif len(p[p!=0]) == 2:
				b = np.zeros(3)
				b[z_idx] = 1.
				b[nz_idx[0]] = a[z_idx]/(a[nz_idx[1]]*p[nz_idx[0]]/p[nz_idx[1]] - a[nz_idx[0]])
				b[nz_idx[1]] = -p[nz_idx[0]]*b[nz_idx[0]]/p[nz_idx[1]]
			else:
				b = np.zeros(3)
				b[nz_idx] = 0
				b[z_idx[0]] = 1.
				b[z_idx[1]] = -a[z_idx[0]]/a[z_idx[1]]
			
			# Convert to unit vectors
			a = a/np.linalg.norm(a)
			b = b/np.linalg.norm(b)
			theta_step = np.pi*2/num_pts
			
			# Define set of points at a defined radius around
			# the original points, C1 and C2
			P1 = np.zeros((num_pts,3))
			P2 = np.zeros((num_pts,3))
			r1 = diams[kk][0]
			r2 = diams[kk][1]
			theta = 0
			for ii in range(num_pts):
				for jj in range(3):
					P1[ii][jj] = C1[jj] + r1*np.cos(theta)*a[jj] + r1*np.sin(theta)*b[jj]
					P2[ii][jj] = C2[jj] + r2*np.cos(theta)*a[jj] + r2*np.sin(theta)*b[jj]
				
				theta += theta_step
			
			# Define triangles
			for ii in range(2*num_pts):
				if ii < num_pts:
					connections.append((ii+offset,(ii+1)%num_pts+offset,ii+num_pts+offset))
				else:
					connections.append((ii+offset,(ii+1-num_pts)%num_pts+offset+num_pts,(ii-num_pts+1)%num_pts+offset))
			
			for ii in range(num_pts):
				x.append(P1[ii][0])
				y.append(P1[ii][1])
				z.append(P1[ii][2])
				D.append(data[kk])
			
			for ii in range(num_pts):
				x.append(P2[ii][0])
				y.append(P2[ii][1])
				z.append(P2[ii][2])
				D.append(data[kk])
			
			offset += 2*num_pts
		
		x = np.array(x)
		y = np.array(y)
		z = np.array(z)
		D = np.array(D)
		
		return x, y, z, connections, D
	
	
	def segment_branch(self,branch):
		segments =[]
		for i,seg_end in enumerate(branch[:-1]):
			segments.append([branch[i],branch[i+1]])
		return(segments)
	
	def unzip_sectioned_arbor(self,arbor):
		if arbor is None:
			return({},{},{},{})
		
		x = {}
		y = {}
		z = {}
		r = {}
		for branch in arbor.keys():
			x[branch] = []
			y[branch] = []
			z[branch] = []
			r[branch] = []
			for section in arbor[branch]:
				for point in section:
					x[branch].append(point[0])
					y[branch].append(point[1])
					z[branch].append(point[2])
					r[branch].append(point[3])
		
		return(x,y,z,r)
	
	def rgb_to_mlabcolor(self,rgb):
		return((rgb[0]/255.0,rgb[1]/255.0,rgb[2]/255.0))
	
	def mplot_sectioned_arbor_simplified(self,arbor,arbor_labels,view=True,DefaultDiameters=True):
		fig = mlab.figure(bgcolor=(42/255.0,56/255.0,54/255.0),size=(1280,720))
		keys = ['node','paranode1','paranode2','internode','interbouton','bouton']
		diams = [0.75,1.54,1.54,1.54,0.2,1.0]
		values = [self.rgb_to_mlabcolor(item) for item in [(255, 22, 84),(112, 193, 179),(178, 219, 191),(36, 123, 160),((243, 255, 189)),(255, 22, 84)]]
		color_dict = dict(zip(keys,values))
		diam_dict = dict(zip(keys,diams))
		mobjs = []
		for branch in arbor.keys():
#			if branch not in [0,1]:
#				continue
			
			for s,section in enumerate(arbor[branch]):
				if DefaultDiameters:
					mobjs.append(mlab.plot3d([sec[0] for sec in section],[sec[1] for sec in section],[sec[2] for sec in section],color=color_dict[arbor_labels[branch][s]],tube_radius=diam_dict[arbor_labels[branch][s]],tube_sides=6,representation='wireframe'))
				else:
					mobjs.append(mlab.plot3d([sec[0] for sec in section],[sec[1] for sec in section],[sec[2] for sec in section],color=color_dict[arbor_labels[branch][s]],tube_radius=section[-1][-1],tube_sides=6))
				
				mobjs[-1].actor.property.lighting = False
		
		mlab.view(azimuth=0,elevation=0)
		
		if view:
			mlab.show()
			
	
	def plot_electrode(self,arbor,arbor_labels,view=False):
		keys = ['contact','noncontact','activecontact']
		values = [self.rgb_to_mlabcolor(item) for item in [(42,56,54),(224, 224, 224),(173,42,42)]]
		color_dict = dict(zip(keys,values))
		electrode_parts = []
		electrode_parts.append(mlab.points3d([arbor[1][0][0][0]],[arbor[1][0][0][1]],[arbor[1][0][0][2]],color=color_dict['noncontact'],scale_factor=arbor[1][0][0][3]*1,mode='sphere',resolution=16))
		for s,section in enumerate(arbor[0]):
			if s in arbor_labels:
				col = color_dict['contact']
				if s == 3:
					col = color_dict['activecontact']
			
			else:
				col = color_dict['noncontact']
			
			electrode_parts.append(mlab.plot3d([sec[0] for sec in section],[sec[1] for sec in section],[sec[2] for sec in section],color=col,tube_radius=section[-1][-1]/2.0,tube_sides=16))
		
		for part in electrode_parts:
			part.actor.property.backface_culling=True
			part.actor.property.frontface_culling=True
			part.actor.property.shading=True
		if view:
			mlab.show()
	
	
	
	def mplot_sectioned_arbors(self,arbors,colors = [(0.29, 0.58, 0.67),(0.82, 0.35, 0.24)],view=True):
		fig = mlab.figure(bgcolor=(42/255.0,56/255.0,54/255.0),size=(1280,720))
		colors = [(item[0]/255.0,item[1]/255.0,item[2]/255.0) for item in [[0,119,187],[51,187,238],[0,153,136],[238,119,51],[204,51,17],[238,51,119],[221,170,51]]]
		colors.reverse()
		col_index = 0
		for arbor in arbors:
			myav_coords = []
			myav_diams = []
			x,y,z,r = self.unzip_sectioned_arbor(arbor)
			coords = []
			diams = []
			for bnum in x:
				tcoords = []
				tdiams = []
				for i,tem in enumerate(x[bnum]):
					tcoords.append([x[bnum][i],y[bnum][i],z[bnum][i]])
					tdiams.append(r[bnum][i])
					tdiams[-1] *= 3.0
				
				coords.extend(self.segment_branch(tcoords))
				diams.extend(self.segment_branch(tdiams))
			
			
			myav_coords.extend(coords)
			myav_diams.extend(diams)
			myav_vs = [20 for i in range(len(myav_coords)-len(coords))]+[2 for j in range(len(coords))]
			num_pts = 20
			tx,ty,tz,tconn,tD = self.create_cylinders(myav_coords,myav_diams,myav_vs,num_pts)
			tmsh = mlab.triangular_mesh(tx,ty,tz,tconn,scalars=tD,vmin=1,vmax=20,representation='wireframe',color=colors[col_index])
			tmsh.actor.property.frontface_culling = True
			tmsh.actor.property.backface_culling = True
			tmsh.actor.property.lighting = False
			col_index+=1
			if col_index==len(colors):
				col_index=0
		
		mlab.view(azimuth=0,elevation=0)
		#	for ii in range(D.shape[1]):
		#		_=mlab.triangular_mesh(x,y,z,connection,scalars = D[:,ii],vmin=Min,vmax=Max)
		#		_=mlab.view(azimuth=0,elevation=0)
		#		_=mlab.savefig('pic%.4d.png' % ii, size=(800,600))
		#	mlab.savefig('pic%.4d.png' % tstep,size=(1200,900))
		
		if view:
			mlab.show()
	
	def view(self):
		mlab.show()
	
	def close(self):
		mlab.close(all=True)
	
	def mplot_sectioned_arbor(self,fig=None,arbor=None,colors = [(0.29, 0.58, 0.67),(0.82, 0.35, 0.24)],view=True):
		if fig is None:
			fig = mlab.figure(bgcolor=(42/255.0,56/255.0,54/255.0),size=(1280,720))
		
		colorind = 0
		myav_coords = []
		myav_diams = []
		x,y,z,r = self.unzip_sectioned_arbor(arbor)
		coords = []
		diams = []
		for bnum in x:
			tcoords = []
			tdiams = []
			for i,tem in enumerate(x[bnum]):
				tcoords.append([x[bnum][i],y[bnum][i],z[bnum][i]])
				tdiams.append(r[bnum][i])
	#			tdiams[-1] = 0.025
			
			coords.extend(self.segment_branch(tcoords))
			diams.extend(self.segment_branch(tdiams))
		
		
		myav_coords.extend(coords)
		myav_diams.extend(diams)
		myav_vs = [20 for i in range(len(myav_coords)-len(coords))]+[2 for j in range(len(coords))]
		num_pts = 20
		tx,ty,tz,tconn,tD = self.create_cylinders(myav_coords,myav_diams,myav_vs,num_pts)
		mlab.triangular_mesh(tx,ty,tz,tconn,scalars=tD,vmin=1,vmax=20)
		colorind+=1
		mlab.view(azimuth=0,elevation=0)
	#	for ii in range(D.shape[1]):
	#		_=mlab.triangular_mesh(x,y,z,connection,scalars = D[:,ii],vmin=Min,vmax=Max)
	#		_=mlab.view(azimuth=0,elevation=0)
	#		_=mlab.savefig('pic%.4d.png' % ii, size=(800,600))
	#	mlab.savefig('pic%.4d.png' % tstep,size=(1200,900))
		if view:
			mlab.show()
	
	
	def mplot_mfile(self,swcfile,colors = [(0.29, 0.58, 0.67),(0.82, 0.35, 0.24)]):
		colorind = 0
		myav_coords = []
		myav_diams = []
		x,y,z,r = self.swcTool.load_swc(swcfile,asTree=False)
		coords = []
		diams = []
		for bnum in x:
			tcoords = []
			tdiams = []
			for i,tem in enumerate(x[bnum]):
				tcoords.append([x[bnum][i],y[bnum][i],z[bnum][i]])
				tdiams.append(r[bnum][i])
	#			tdiams[-1] = 0.025
			
			coords.extend(self.segment_branch(tcoords))
			diams.extend(self.segment_branch(tdiams))
		
		
		myav_coords.extend(coords)
		myav_diams.extend(diams)
		myav_vs = [20 for i in range(len(myav_coords)-len(coords))]+[2 for j in range(len(coords))]
		num_pts = 6
		tx,ty,tz,tconn,tD = self.create_cylinders(myav_coords,myav_diams,myav_vs,num_pts)
		mlab.triangular_mesh(tx,ty,tz,tconn,scalars=tD,vmin=1,vmax=20,color=colors[colorind])
		colorind+=1
		mlab.view(azimuth=0,elevation=0)
	#	for ii in range(D.shape[1]):
	#		_=mlab.triangular_mesh(x,y,z,connection,scalars = D[:,ii],vmin=Min,vmax=Max)
	#		_=mlab.view(azimuth=0,elevation=0)
	#		_=mlab.savefig('pic%.4d.png' % ii, size=(800,600))
	#	mlab.savefig('pic%.4d.png' % tstep,size=(1200,900))
		mlab.show()







